from typing import Optional

from asset_discovery.asset_cache.asset_cache_manager import AssetCacheManager
from asset_discovery.asset_cache.inner_cache_config import InnerCacheConfig
from asset_discovery.asset_cache.settings_provider import SettingsProvider
from asset_discovery.exceptions import DuplicateError, NotFoundError, NotSupportedError, OperationError
from asset_discovery.models import RefreshReason, OperationResult, OperationStatus, ReportingStatus, Asset
from asset_discovery.notification.callback_registry import CallbackRegistry
from asset_discovery.util.plugin_manager import PluginManager
from asset_discovery.util.plugin_manager_base import PluginManagerBase
from asset_discovery.util.named_mutex import NamedMutex
from asset_discovery.util.named_mutex_exception import NamedMutexTimeoutException
import logging

import os
if os.name == 'nt':
    from asset_discovery.file_monitor.file_monitor import FileMonitor
else:
    from asset_discovery.file_monitor.linux_file_monitor import LinuxFileMonitor as FileMonitor

logger = logging.getLogger()


class DiscoveryEngine:
    _DISCOVERY_ENGINE_MUTEX_NAME = "asset_discovery.discovery_engine"

    def __init__(self, plugin_manager: PluginManagerBase = None, asset_cache_manager: AssetCacheManager = None):
        if plugin_manager is None:
            plugin_manager = PluginManager()
        self._plugin_manager = plugin_manager
        if asset_cache_manager is None:
            asset_cache_manager = AssetCacheManager()
        self._asset_cache_manager = asset_cache_manager
        self._configure_caches()
        self._callback_registry = CallbackRegistry()
        self._register_plugins_callback()
        settings_provider = SettingsProvider()
        merged_cache_store_path = settings_provider.get_merged_store_directory_path()
        self._monitor = FileMonitor(
            merged_cache_store_path,
            file_name="merged_cache_last_update_timestamp.json",
            callback_registry=self._callback_registry)

    def __del__(self):
        self._monitor.stop_monitoring()

    @staticmethod
    def _log_error(e):
        logger.error(e)

    def _configure_caches(self):
        plugins = self._plugin_manager.get_sorted_plugins(high_to_low=False)
        plugin_cache_configs = [InnerCacheConfig(plugin_id, plugin.merge) for plugin_id, plugin in plugins.items()]
        self._asset_cache_manager.configure_inner_caches(plugin_cache_configs)
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                self._asset_cache_manager.recompute_merged_cache()
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    def _register_plugins_callback(self):
        plugins = self._plugin_manager.get_sorted_plugins(high_to_low=False)
        for plugin_id, plugin in plugins.items():
            # send the plugin_id as callback id to ensure no callback duplication
            plugin.register_change_callback(plugin_id, self._plugin_on_change_callback)

    def _plugin_on_change_callback(self, plugin_id):
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                self._refresh_plugin_cache(plugin_id, RefreshReason.INFRASTRUCTURE_REQUEST)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    def _refresh_plugin_cache(self, plugin_id, refresh_reason, recompute=True, **kwargs):
        plugin = self._plugin_manager.get_plugin_by_id(plugin_id)
        # noinspection PyBroadException
        try:
            plugin_assets = plugin.list_assets(refresh_reason=refresh_reason, **kwargs)
            if plugin_assets is not None:  # TODO: Update this check to verify that the operations was successful
                self._asset_cache_manager.reset_cache(plugin_id, plugin_assets, recompute)
                return True
        except NotSupportedError as e:
            pass  # The operation is not supported on the plugin, continue
        except OperationError as e:
            self._log_error(e)
        except Exception as e:
            self._log_error(e)
        return False

    def register_change_callback(self, callback_id, callback, *args):
        self._callback_registry.register_change_callback(callback_id, callback, *args)
        self._start_monitoring()

    def _start_monitoring(self):
        return self._monitor.start_monitoring()

    def list_assets(self, **kwargs):
        include_unreported = kwargs.get("include_unreported", False)
        assets = []
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                assets = self._asset_cache_manager.get_all(include_unreported)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

        return assets

    def get_by_identification(self, asset_identification, **kwargs) -> Optional[Asset]:
        asset = None
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                asset = self._asset_cache_manager.get_by_identification(asset_identification)
        except NamedMutexTimeoutException as e:
            self._log_error(e)
        return asset

    def add_asset(self, asset, **kwargs):
        plugin_id = kwargs.get("plugin", None)
        if plugin_id is None or plugin_id not in self._plugin_manager.plugins.keys():
            plugin_id = "json"

        plugin = self._plugin_manager.plugins[plugin_id]
        # noinspection PyBroadException
        try:
            plugin.add_asset(asset, **kwargs)
        except NotSupportedError:
            json_plugin = self._plugin_manager.plugins["json"]
            # noinspection PyBroadException
            try:
                json_plugin.add_asset(asset)
            except DuplicateError as e:
                self._log_error(e)
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)
        except DuplicateError as e:
            self._log_error(e)
        except OperationError as e:
            self._log_error(e)
        except Exception as e:
            self._log_error(e)

    def remove_asset(self, asset_identification, **kwargs):
        plugins = self._plugin_manager.get_sorted_plugins()
        for _, plugin in plugins.items():
            # noinspection PyBroadException
            try:
                plugin.remove_asset(asset_identification, **kwargs)
            except NotSupportedError:
                continue  # The operation is not supported on the plugin, continue
            except NotFoundError:
                continue  # The resource is not handled by the plugin, continue
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)

    def update_asset(self, asset_update, **kwargs) -> None:
        plugins = self._plugin_manager.get_sorted_plugins()
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                metadata = self._asset_cache_manager.get_asset_metadata(asset_update.identification)
        except NamedMutexTimeoutException as e:
            self._log_error(e)
            return

        if metadata is None or metadata.reporting_status == ReportingStatus.UNREPORTED_RESOURCE:
            return
        partial_update = metadata.reporting_status == ReportingStatus.MULTI_SOURCE_RESOURCE
        plugins_for_asset = self._get_plugins_for_asset(metadata, plugins)
        for _, plugin in plugins_for_asset.items():
            # noinspection PyBroadException
            try:
                kwargs["partial_update"] = partial_update
                plugin.update_asset(asset_update, **kwargs)
            except NotSupportedError:
                continue  # The operation is not supported on the plugin, continue
            except NotFoundError:
                continue  # The resource is not handled by the plugin, continue
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)

    @staticmethod
    def _get_plugins_for_asset(metadata, plugins):
        return {
            plugin_id: plugin
            for plugin_id, plugin
            in plugins.items()
            if plugin_id in metadata.data_sources
        }

    def self_test(self, asset_identification, **kwargs):
        # kwargs for modes
        plugins = self._plugin_manager.get_sorted_plugins()
        for _, plugin in plugins.items():
            # noinspection PyBroadException
            try:
                result = plugin.self_test(asset_identification, **kwargs)
                return result.status.succeeded
            except NotSupportedError:
                continue  # The operation is not supported on the plugin, continue
            except NotFoundError:
                continue  # The resource is not handled by the plugin, continue
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)
        return False

    def self_calibrate(self, asset_identification, **kwargs):
        # kwargs for modes
        plugins = self._plugin_manager.get_sorted_plugins()
        for _, plugin in plugins.items():
            # noinspection PyBroadException
            try:
                result = plugin.self_calibrate(asset_identification, **kwargs)
                return result.status.succeeded
            except NotSupportedError:
                continue  # The operation is not supported on the plugin, continue
            except NotFoundError:
                continue  # The resource is not handled by the plugin, continue
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)
        return False

    def reset(self, asset_identification, **kwargs):
        # kwargs for modes
        plugins = self._plugin_manager.get_sorted_plugins()
        for _, plugin in plugins.items():
            # noinspection PyBroadException
            try:
                result = plugin.reset(asset_identification, **kwargs)
                return result.status.succeeded
            except NotSupportedError:
                continue  # The operation is not supported on the plugin, continue
            except NotFoundError:
                continue  # The resource is not handled by the plugin, continue
            except OperationError as e:
                self._log_error(e)
            except Exception as e:
                self._log_error(e)
        return False

    def refresh(self, **kwargs):
        # kwargs for sys api attributes
        operation_status = 0
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                plugins = self._plugin_manager.get_sorted_plugins(high_to_low=False)
                any_changes = False
                for plugin_id, _ in plugins.items():
                    # noinspection PyBroadException
                    try:
                        any_changes_for_plugin = self._refresh_plugin_cache(
                            plugin_id,
                            RefreshReason.USER_REQUEST,
                            False,
                            **kwargs)
                        any_changes = any_changes or any_changes_for_plugin
                    except NotSupportedError:
                        continue  # The operation is not supported on the plugin, continue
                    except OperationError as e:
                        self._log_error(e)
                    except Exception as e:
                        self._log_error(e)

                if any_changes:
                    self._asset_cache_manager.recompute_merged_cache()
        except NamedMutexTimeoutException as e:
            self._log_error(e)

        return operation_status

    def get_metadata(self, **kwargs):
        # noinspection PyBroadException
        try:
            with NamedMutex(DiscoveryEngine._DISCOVERY_ENGINE_MUTEX_NAME):
                metadata = self._asset_cache_manager.get_all_metadata()
            status = OperationStatus(True, 0)
            return OperationResult(status, metadata)
        except Exception as e:
            self._log_error(e)
            status = OperationStatus(False, -1)
            return OperationResult(status)
