import threading
import time
import logging
from typing import Dict, List, Union

from asset_discovery.exceptions import NotFoundError, OperationError
from asset_discovery.models import (
    Asset,
    AssetUpdate,
    AssetPresenceStatus,
    DiscoveryType,
    OperationResult,
    OperationStatus,
    RefreshReason, AssetIdentification)
from asset_discovery.models.delete_mode import DeleteMode
from asset_discovery.plugins.sysapi_imports.repository.bag_serializer import BagSerializer
from asset_discovery.plugins.sysapi_imports.repository.ipc_bag_provider import IPCBagProvider
from asset_discovery.plugins.sysapi_imports.repository.models import PropertyBag, PropertyBagRequest
from asset_discovery.notification.callback_registry import CallbackRegistry
from asset_discovery.plugins.sysapi_imports.repository.models.asset_resource_update import AssetResourceUpdate
from asset_discovery.plugins.sysapi_imports.repository.fan_sensor_serializer import FanSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.power_sensor_serializer import PowerSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.temperature_sensor_serializer import TemperatureSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.settings_provider import SettingsProvider
from asset_discovery.plugins.sysapi_imports.repository.overwritten_properties import OverwrittenProperties
from asset_discovery.plugins.sysapi_imports.repository.voltage_sensor_serializer import VoltageSensorSerializer
from asset_discovery.plugins.sysapi_imports.request_helpers.add.add_request_builder import AddRequestBuilder
from asset_discovery.repository.file_json_provider import FileJsonProvider
from asset_discovery.repository.json_data_store import JsonDataStore
from asset_discovery.util.asset_serializer import AssetSerializer
from asset_discovery.util.named_mutex import NamedMutex
from asset_discovery.util.named_mutex_exception import NamedMutexTimeoutException
from asset_discovery.util.object_patcher import ObjectPatcher
from asset_discovery.models import Defaults

logger = logging.getLogger()

DEFAULT = Defaults.DEFAULT_VALUE


class AssetPlugin:
    _SYSAPI_MUTEX_NAME = "asset_discovery.plugins.sysapi"

    def __init__(self, bag_provider=None, json_provider=None):
        settings_provider = SettingsProvider()
        store_directory_path = settings_provider.get_store_directory_path()
        if json_provider is None:
            json_provider = FileJsonProvider(store_directory_path)
        self._overwritten_props_data_store = JsonDataStore(
            json_provider,
            AssetSerializer(),
            top_level_property_name="overwritten_props")
        if bag_provider is None:
            bag_provider = IPCBagProvider()
        self._bag_provider = bag_provider
        self._bag_serializer = BagSerializer()
        self._fan_sensor_serializer = FanSensorSerializer()
        self._temperature_sensor_serializer = TemperatureSensorSerializer()
        self._voltage_sensor_serializer = VoltageSensorSerializer()
        self._power_sensor_serializer = PowerSensorSerializer()
        self._callback_registry = CallbackRegistry()
        self._monitoring_thread = None
        self._should_monitor = False
        self._overwritten_props_dirty = False
        self._add_request_helper = AddRequestBuilder()

    def __del__(self):
        self._should_monitor = False
        if self._monitoring_thread is not None:
            self._monitoring_thread.join()
            self._monitoring_thread = None

    @staticmethod
    def _log_error(e):
        logger.error(e)

    def register_change_callback(self, callback_id, callback, *args):
        plugin_id = __name__.split('.')[-1]
        self._callback_registry.register_change_callback(callback_id, callback, plugin_id)
        self._monitoring_thread = threading.Thread(target=self._start_monitoring, daemon=True)
        self._monitoring_thread.start()

    def _start_monitoring(self):
        self._should_monitor = True
        while self._should_monitor:
            self._check_for_updates()
            time.sleep(2)

    def _check_for_updates(self):
        if self._overwritten_props_dirty or (self._bag_provider.any_properties_changed() is True):
            self._overwritten_props_dirty = False
            self._callback_registry.call()

    def list_assets(self, refresh_reason=RefreshReason.INFRASTRUCTURE_REQUEST, **kwargs) -> List[Asset]:
        # Always get all safe asset data
        all_bags = self._bag_provider.get_all()

        # Identify assets we want to retrieve unsafe data for
        requested_assets = kwargs.get("assets", [])
        filtered_bags = self._filter_bags_to_populate(all_bags, refresh_reason, requested_assets)

        # Populate unsafe asset data
        requested_bags = self._get_requested_bags(filtered_bags)
        populated_bags = []
        if requested_bags:
            populated_bags = self._bag_provider.get_unsafe(requested_bags)

        # Cache some populated data in overwritten props
        self._cache_populated_data(filtered_bags, populated_bags)

        # Get merged assets (safe data + unsafe data)
        merged_bags = self._get_merged_bags(all_bags, populated_bags)
        merged_assets = self._get_assets_from_bags(merged_bags)

        # Update assets with data from overwritten props
        overwritten_props = self._overwritten_props_data_store.get_identification_to_entry()
        overwritten_assets = self._merge_overwritten_props(merged_assets, overwritten_props)

        return overwritten_assets

    def _filter_bags_to_populate(self, all_bags, refresh_reason, requested_assets):
        if refresh_reason == RefreshReason.USER_REQUEST:
            return self._filter_bags_to_populate_on_user_request(all_bags, requested_assets)
        else:
            return []

    def _filter_bags_to_populate_on_user_request(self, all_bags, requested_assets):
        if not requested_assets:
            return all_bags

        bags_to_refresh = []
        possible_parents = self._bag_serializer.get_possible_parents_from_bags(all_bags)
        identification_models = self._create_identification_models_from_dictionaries(requested_assets)
        for bag in all_bags:
            asset = self._bag_serializer.from_bag(bag, possible_parents)
            if asset is None:
                continue
            if asset.identification in identification_models or self._bag_serializer.is_controller_or_chassis(bag):
                bags_to_refresh.append(bag)
        return bags_to_refresh

    @staticmethod
    def _create_identification_models_from_dictionaries(dictionaries: List[Dict[str, str]]) -> List[AssetIdentification]:
        return [AssetPlugin._create_identification_model_from_dictionary(dictionary) for dictionary in dictionaries]

    @staticmethod
    def _create_identification_model_from_dictionary(dictionary: Dict[str, Union[int, str]]) -> AssetIdentification:
        model_name = dictionary.get("model_name", "")
        model_number = dictionary.get("model_number", 0)
        vendor_name = dictionary.get("vendor_name", "")
        vendor_number = dictionary.get("vendor_number", 0)
        serial_number = dictionary.get("serial_number", "")
        return AssetIdentification(
            model_name=model_name,
            model_number=model_number,
            vendor_name=vendor_name,
            vendor_number=vendor_number,
            serial_number=serial_number
        )

    def _get_requested_bags(self, bags) -> List[PropertyBagRequest]:
        return [self._create_standard_and_health_bag_request(bag.resource_uri) for bag in bags]

    @staticmethod
    def _create_standard_and_health_bag_request(resource_uri) -> PropertyBagRequest:
        requested_bag = PropertyBagRequest(resource_uri)
        requested_bag.add_standard_properties()
        requested_bag.add_health_properties()
        return requested_bag

    def _cache_populated_data(self, filtered_bags, populated_bags) -> None:
        merged_populated_bags = self._get_merged_bags(filtered_bags, populated_bags)
        merged_populated_assets = self._get_assets_from_bags(merged_populated_bags)
        overwritten_props = self._overwritten_props_data_store.get_identification_to_entry()
        for asset in merged_populated_assets:
            asset_update, should_update = self._get_overwritten_props_update_from_populated_data(
                asset, overwritten_props)
            if should_update:
                self._overwrite_sysapi_props(asset_update)

    @staticmethod
    def _get_merged_bags(all_bags, populated_bags) -> List[PropertyBag]:
        resource_uri_to_props = {bag.resource_uri: bag.properties for bag in populated_bags}
        merged_bags = []
        for bag in all_bags:
            uri = bag.resource_uri
            properties = bag.properties
            properties.update(resource_uri_to_props.get(uri, {}))
            merged_bags.append(PropertyBag(uri, properties))
        return merged_bags

    def _get_assets_from_bags(self, merged_bags) -> List[Asset]:
        possible_parents = self._bag_serializer.get_possible_parents_from_bags(merged_bags)
        assets = {}
        for bag in merged_bags:
            asset = self._bag_serializer.from_bag(bag, possible_parents)
            if asset is not None:
                asset.discovery_type = DiscoveryType.AUTOMATIC
                asset_presence = asset.location.state.asset_presence
                if asset.identification in assets and asset_presence == AssetPresenceStatus.NOT_PRESENT:
                    continue  # Asset already reported, and newer entry is Not Present, so we don't overwrite it
                assets[asset.identification] = asset
        return list(assets.values())

    @staticmethod
    def _get_overwritten_props_update_from_populated_data(asset, overwritten_props):
        asset_update = AssetUpdate.create_from_identification(asset.identification)
        should_update = False
        asset_health = asset.health
        if asset_health is not None:
            asset_update.health = asset_health
            should_update = True
        external_calibration = asset.external_calibration
        if external_calibration is not None:
            asset_overwritten_props = overwritten_props.get(asset.identification, None)
            if asset_overwritten_props is not None:
                existing_external_calibration = asset_overwritten_props.external_calibration
                if existing_external_calibration is not None \
                        and existing_external_calibration.date < external_calibration.date:
                    asset_update.external_calibration = None
                    should_update = True
        return asset_update, should_update

    @staticmethod
    def _merge_overwritten_props(sys_api_assets, overwritten_props) -> List[Asset]:
        merged_assets = []
        for asset in sys_api_assets:
            asset_overwritten_props = overwritten_props.get(asset.identification)
            if asset_overwritten_props:
                merged_assets.append(ObjectPatcher.patch(asset, asset_overwritten_props))
            else:
                merged_assets.append(asset)
        return merged_assets

    def self_test(self, asset_identification, **kwargs):
        mode = kwargs.get('mode', 0)

        bag_matching_identification = self._get_bag_with_identification(asset_identification)
        if bag_matching_identification is None:
            raise NotFoundError(asset_identification)

        result = self._bag_provider.self_test(bag_matching_identification, mode)
        status = result.get('status_code', 0)
        if status != 0:
            return OperationResult(OperationStatus(False, status, 'Self test operation failed on the SysAPI plugin'))
        return OperationResult(OperationStatus(True))

    def self_calibrate(self, asset_identification, **kwargs):
        bag_matching_identification = self._get_bag_with_identification(asset_identification)
        if bag_matching_identification is None:
            raise NotFoundError(asset_identification)
        result = self._bag_provider.self_calibrate(bag_matching_identification)
        status = result.get('status_code', 0)
        if status != 0:
            return OperationResult(
                OperationStatus(False, status, 'Self calibrate operation failed on the SysAPI plugin'))
        return OperationResult(OperationStatus(True))

    def reset(self, asset_identification, **kwargs):
        mode = kwargs.get("mode", 0)
        bag_matching_identification = self._get_bag_with_identification(asset_identification)
        status = OperationStatus(successful=True)

        # If we found a single bag with the given identification, use that
        if bag_matching_identification is None:
            raise NotFoundError(asset_identification)

        result = self._bag_provider.self_reset(bag_matching_identification, mode)

        result_status = result.get('status_code', 0)
        if result_status != 0:
            status = OperationStatus(False, result_status, "Reset operation did not succeed on the SysAPI Plugin")
        return OperationResult(status=status)

    def _get_bag_with_identification(self, asset_identification):
        bags = self._bag_provider.get_all()
        matched_bags = [a for a in bags if self._is_bag_matching_asset_identification(a, asset_identification)]

        matched_bags_count = len(matched_bags)
        if matched_bags_count == 0:
            return None
        if matched_bags_count == 1:
            # If we found a single bag with the given identification, use that
            return matched_bags[0]
        else:
            for bag in matched_bags:
                if self._bag_serializer.is_present(bag):
                    # If we found multiple bags with the given identification, use the Present one
                    return bag
            # If we found multiple bags with the given identification, and none is Present, use any one
            return matched_bags[0]

    def _is_bag_matching_asset_identification(self, bag, asset_identification):
        possible_parents = self._bag_serializer.get_possible_parents_from_bags([bag])
        asset = self._bag_serializer.from_bag(bag, possible_parents)
        if asset is not None and asset.identification == asset_identification:
            return True
        return False

    def update_asset(self, asset_update, **kwargs):
        bag_matching_identification = self._get_bag_with_identification(asset_update.identification)
        if bag_matching_identification is None:
            raise NotFoundError(asset_update.identification)

        request = AssetResourceUpdate(bag_matching_identification.resource_uri)
        result = None
        if asset_update.name is not DEFAULT:
            request.set_updated_name(asset_update.name)
            result = self._bag_provider.update(request)

        self._overwrite_sysapi_props(asset_update)

        status = OperationStatus(successful=True)
        if result:
            result_status = result.get('status_code', 0)
            if result_status != 0:
                status = OperationStatus(False, result_status, "Update operation did not succeed on the SysAPI Plugin")
        return OperationResult(status=status)

    def _overwrite_sysapi_props(self, asset_update):
        try:
            with NamedMutex(AssetPlugin._SYSAPI_MUTEX_NAME):
                try:
                    overwritten_asset_update, any_overwritten_prop = self._get_overwritten_asset_update(asset_update)
                    self._overwritten_props_data_store.update(overwritten_asset_update)
                    self._overwritten_props_dirty = any_overwritten_prop   # TODO
                except NotFoundError:
                    asset = self._get_overwritten_asset(asset_update)
                    self._overwritten_props_data_store.add(asset)
                    self._overwritten_props_dirty = True
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    @staticmethod
    def _get_overwritten_asset_update(asset_update):
        overwritten_asset_update = AssetUpdate.create_from_identification(asset_update.identification)
        any_overwritten_prop = False
        for getter, setter in OverwrittenProperties.get_overwritten_properties():
            prop_value = getter(asset_update)
            if prop_value is not DEFAULT:
                setter(overwritten_asset_update, prop_value)
                any_overwritten_prop = True
        return overwritten_asset_update, any_overwritten_prop

    @staticmethod
    def _get_overwritten_asset(asset_update):
        overwritten_asset = Asset.create_from_identification(asset_update.identification)
        for getter, setter in OverwrittenProperties.get_overwritten_properties():
            prop_value = getter(asset_update)
            if prop_value is not DEFAULT:
                setter(overwritten_asset, prop_value)
        return overwritten_asset

    def add_asset(self, asset: Asset, **kwargs):
        build_request_result = self._add_request_helper.build_from(asset, **kwargs)
        if not build_request_result.status.succeeded:
            logger.error(build_request_result.status.description)
            return build_request_result
        request = build_request_result.content

        create_resource_result = self._bag_provider.create(request_kwargs=request)
        if not create_resource_result.status.succeeded:
            error_code = str(create_resource_result.status.code)
            message = "Create resource request failed with error code " + error_code + "."
            logger.error(message)
        return create_resource_result

    def remove_asset(self, asset_identification, **kwargs):
        try:
            with NamedMutex(AssetPlugin._SYSAPI_MUTEX_NAME):
                mode = kwargs.get('mode', DeleteMode.LEAVE_DEPENDENT)
                sysapi_mode = self._convert_delete_mode(mode)
                bag_matching_identification = self._get_bag_with_identification(asset_identification)
                if bag_matching_identification is None:
                    raise NotFoundError(asset_identification)
                if not self._bag_serializer.can_delete(bag_matching_identification):
                    message = "Asset matched by identification {0} cannot be deleted.".format(str(asset_identification))
                    raise OperationError(message=message)
                resource_uri = bag_matching_identification.resource_uri
                self._bag_provider.remove(resource_uri=resource_uri, mode=sysapi_mode)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    @staticmethod
    def _convert_delete_mode(mode: str) -> int:
        if mode == DeleteMode.IF_NO_DEPENDENT:
            return 0
        if mode == DeleteMode.WITH_DEPENDENT:
            return 1
        return 2

"""
What tags are sent on refresh (populate_props) from D&I:
                  16969728,FirmwareRevision
                  16965632,SensorTempReading0
                  16961536,CalMiscComments
                  17428480,CalExtLastLimited
                  16867328,CalExtLastTempC
                  16863232,CalExtLastTime
                  16871424,CalExtNextTime
                  17207296,CalExtRecommendedInterval
                  17174528,SensorFanCount
                  17256448,HardwareRevision
                  17420288,CalIntLastLimited
                  16850944,CalIntLastTempC
                  16846848,CalIntLastTime
                  17178624,SensorFanName0
                  17182720,SensorFanReading0
                  17190912,SensorTempName0
                  17154048,SensorVoltName0
                  17162240,SensorVoltNominal0
                  17158144,SensorVoltReading0
                  17186816,SensorTempCount
                  17149952,SensorVoltCount
"""
