from datetime import datetime
from typing import Optional, List

from asset_discovery.models import FanSensorModel, CoolingHealth, CoolingProfile, FanMode, OperationalStateValue
from asset_discovery.plugins.sysapi_imports.repository.models import (
    AttributeTag,
    PropertyBag,
    AttributeCoolingProfile,
    AttributeFanMode)
from asset_discovery.plugins.sysapi_imports.repository.base_sensor_serializer import AbstractSensorSerializer


class FanSensorSerializer:
    COOLING_PROFILE_VALUES = {
        AttributeCoolingProfile.Profile38W: CoolingProfile.PROFILE_38_W,
        AttributeCoolingProfile.Profile58W: CoolingProfile.PROFILE_58_W,
        AttributeCoolingProfile.Profile82W: CoolingProfile.PROFILE_82_W,
        AttributeCoolingProfile.ProfileCustom: CoolingProfile.PROFILE_CUSTOM
    }

    FAN_MODE_VALUES = {
        AttributeFanMode.Auto: FanMode.AUTO,
        AttributeFanMode.SafeManual: FanMode.SAFE_MANUAL,
        AttributeFanMode.High: FanMode.HIGH
    }

    OPERATIONAL_STATE_VALUES = {
        0: OperationalStateValue.UNKNOWN,
        1: OperationalStateValue.NORMAL,
        256: OperationalStateValue.DEGRADED,
        512: OperationalStateValue.CONFIG_WARNING,
        1024: OperationalStateValue.PREDICTIVE_FAILURE,
        65536: OperationalStateValue.RECOVERABLE_ERROR_PROGRAMMATIC,
        131072: OperationalStateValue.RECOVERABLE_ERROR_USER_ACTION,
        262144: OperationalStateValue.NON_RECOVERABLE_ERROR
    }

    def __init__(self):
        self._base_fan_name = "Fan"

    def serialize(self, bag: PropertyBag, timestamp: datetime) -> Optional[CoolingHealth]:
        fan_sensors = self._create_fan_sensors(bag)
        is_chassis = FanSensorSerializer._is_chassis(bag)
        power_fan_sensors = self._create_power_supply_fan_sensors(bag) if is_chassis else None
        fan_mode = self._create_fan_mode(bag) if is_chassis else None
        supported_fan_modes = self._create_supported_fan_modes(bag) if is_chassis else None
        user_desired_fan_speed = self._create_fan_user_rpm(bag) if is_chassis and fan_mode == FanMode.SAFE_MANUAL else None
        manual_rpm_upper_bound = self._create_manual_rpm_upper_bound(bag) if is_chassis and fan_mode == FanMode.SAFE_MANUAL else None
        manual_rpm_lower_bound = self._create_manual_rpm_lower_bound(bag) if is_chassis and fan_mode == FanMode.SAFE_MANUAL else None
        cooling_profile = self._create_cooling_profile(bag) if is_chassis else None
        supported_cooling_profiles = self._create_supported_cooling_profiles(bag) if is_chassis else None
        cooling_operational_state = self._create_cooling_operational_state(bag) if is_chassis else None

        if (not fan_sensors and not power_fan_sensors and not fan_mode and not user_desired_fan_speed and
                not manual_rpm_upper_bound and not manual_rpm_lower_bound and not cooling_profile and
                not supported_cooling_profiles and not cooling_operational_state):
            return None
        return CoolingHealth(fan_mode=fan_mode,
                             supported_fan_modes=supported_fan_modes,
                             user_desired_fan_speed=user_desired_fan_speed,
                             manual_rpm_upper_bound=manual_rpm_upper_bound,
                             manual_rpm_lower_bound=manual_rpm_lower_bound,
                             cooling_profile=cooling_profile,
                             supported_cooling_profiles=supported_cooling_profiles,
                             cooling_operational_state=cooling_operational_state,
                             chassis_fans=fan_sensors,
                             power_supply_fans=power_fan_sensors,
                             timestamp=timestamp)

    @staticmethod
    def _is_chassis(bag):
        return bag.get_property_or_default(AttributeTag.ItemIsChassis, False)

    def _create_fan_mode(self, bag: PropertyBag) -> str:
        fan_mode_attribute_value = bag.get_property_or_default(AttributeTag.FanMode, None)
        fan_mode = self.FAN_MODE_VALUES[fan_mode_attribute_value] if fan_mode_attribute_value is not None else None
        return fan_mode

    def _create_supported_fan_modes(self, bag: PropertyBag) -> Optional[List[str]]:
        supported_fan_modes_attribute_value = bag.get_property_or_default(AttributeTag.SupportedFanModes, None)
        if supported_fan_modes_attribute_value is None:
            return None
        all_possible_fan_modes = [AttributeFanMode.Auto, AttributeFanMode.SafeManual, AttributeFanMode.High]
        supported_fan_modes = []
        for possible_fan_mode in all_possible_fan_modes:
            # based on current fan_mode values, assume fan_mode is always a power of 2
            if (supported_fan_modes_attribute_value & possible_fan_mode) == possible_fan_mode:
                supported_fan_modes.append(self.FAN_MODE_VALUES[possible_fan_mode])

        return supported_fan_modes

    @staticmethod
    def _create_fan_user_rpm(bag: PropertyBag) -> int:
        fan_user_requested_rpm = bag.get_property_or_default(AttributeTag.FanUserRpm, None)
        return fan_user_requested_rpm

    @staticmethod
    def _create_manual_rpm_upper_bound(bag: PropertyBag) -> int:
        manual_rpm_upper_bound = bag.get_property_or_default(AttributeTag.FanManualRpmUpperBound, None)
        return manual_rpm_upper_bound

    @staticmethod
    def _create_manual_rpm_lower_bound(bag: PropertyBag) -> int:
        manual_rpm_lower_bound = bag.get_property_or_default(AttributeTag.FanManualRpmLowerBound, None)
        return manual_rpm_lower_bound

    def _create_cooling_profile(self, bag: PropertyBag) -> str:
        cooling_profile_value = bag.get_property_or_default(AttributeTag.CoolingProfile, None)
        cooling_profile = self.COOLING_PROFILE_VALUES[cooling_profile_value] if cooling_profile_value is not None else None
        return cooling_profile

    def _create_supported_cooling_profiles(self, bag: PropertyBag) -> Optional[List[str]]:
        supported_cooling_profiles_value = bag.get_property_or_default(AttributeTag.SupportedCoolingProfiles, None)
        if supported_cooling_profiles_value is None:
            return None
        all_possible_cooling_profiles = [AttributeCoolingProfile.Profile38W,
                                         AttributeCoolingProfile.Profile58W,
                                         AttributeCoolingProfile.Profile82W,
                                         AttributeCoolingProfile.ProfileCustom]
        supported_cooling_profiles = []
        for possible_cooling_profile in all_possible_cooling_profiles:
            # based on current cooling_profile values, assume cooling_profile is always a power of 2
            if (supported_cooling_profiles_value & possible_cooling_profile) == possible_cooling_profile:
                supported_cooling_profiles.append(self.COOLING_PROFILE_VALUES[possible_cooling_profile])

        return supported_cooling_profiles

    def _create_fan_sensors(self, bag: PropertyBag) -> List[FanSensorModel]:
        fan_sensor_count = bag.get_property_or_default(AttributeTag.SensorFanCount, 0)
        fan_sensors = []
        for i in range(fan_sensor_count):
            sensor = self._fan_sensor_serialize_base(i, bag)
            fan_sensors.append(sensor)
        return fan_sensors

    def _fan_sensor_serialize_base(self, index: int, bag: PropertyBag) -> FanSensorModel:
        sensor = FanSensorModel()
        sensor.name = self._get_name_from_bag(index, bag)
        sensor.reading = bag.get_property_or_default(AttributeTag.SensorFanReading0 + index, None)
        state = bag.get_property_or_default(AttributeTag.SensorFanHealthState0 + index, None)
        sensor.state = AbstractSensorSerializer.SENSOR_HEALTH_STATE_VALUES.get(state, None)

        properties = {}
        # TODO: Add custom sensor properties
        sensor.properties = properties

        return sensor

    def _get_name_from_bag(self, index: int, bag: PropertyBag) -> str:
        default_name = "{0}{1}".format(self._base_fan_name, index)
        return bag.get_property_or_default(AttributeTag.SensorFanName0 + index, default_name)

    def _create_power_supply_fan_sensors(self, bag: PropertyBag) -> Optional[List[FanSensorModel]]:
        power_supply_fan_sensor_count = bag.get_property_or_default(AttributeTag.PowerSupplyFanCount0, 0)
        if not power_supply_fan_sensor_count > 0:
            return None
        power_supply_fan_sensors = []
        for i in range(power_supply_fan_sensor_count):
            sensor = self._power_supply_fan_sensor_serialize_base(i, bag)
            power_supply_fan_sensors.append(sensor)
        return power_supply_fan_sensors

    def _power_supply_fan_sensor_serialize_base(self, index: int, bag: PropertyBag) -> FanSensorModel:
        sensor = FanSensorModel()
        sensor.name = "{0} {1}".format(self._base_fan_name, index)
        sensor.reading = bag.get_property_or_default(AttributeTag.PowerSupplyFanReading0 + index, None)
        return sensor

    def _create_cooling_operational_state(self, bag: PropertyBag) -> str:
        operational_state_value = bag.get_property_or_default(AttributeTag.CoolingOperationalState, None)
        operational_state = self.OPERATIONAL_STATE_VALUES[operational_state_value] if operational_state_value is not None else None
        return operational_state
