from datetime import datetime, timedelta, timezone


class DateTimeUtils:
    minimum_valid_calibration_date = datetime(1976, 1, 1, tzinfo=timezone.utc)

    @staticmethod
    def convert_sys_api_serialized_date_to_date(serialized_date):
        """
        Converts a string containing a System API-specific representation of a date to a datetime object.
        The implementation is adapted from SysTimestampHelper.cs (search under ni-central repo).

        :param serialized_date: The date represented as a string.
        :return: The parsed date in datetime format.
        """
        max_ulong = 0xffffffffffffffff
        offset_between_0001_and_1904 = 0xfffffff20493b980
        ticks_conversions_unit = 10000000

        date_values = serialized_date.split(' ')
        t_1 = int(date_values[0], 16)
        t_2 = int(date_values[1], 16)
        t_3 = int(date_values[2], 16)
        t_4 = int(date_values[3], 16)

        sec_since_epoch_1904 = t_4 << 32 | t_3
        sec_since_epoch_0001 = sec_since_epoch_1904 - offset_between_0001_and_1904
        # System API relies on the rollover property of unsigned int.
        # Since we don't have an unsigned int type readily available,
        # we'll handle the rollover manually.
        if sec_since_epoch_0001 < 0:
            sec_since_epoch_0001 += max_ulong + 1
        frac_int = t_2 << 32 | t_1
        frac_real = 2.0 ** (-64.0) * frac_int

        ticks = sec_since_epoch_0001 * ticks_conversions_unit + frac_real * ticks_conversions_unit
        return datetime(1, 1, 1, tzinfo=timezone.utc) + timedelta(microseconds=ticks / 10)

    @staticmethod
    def convert_date_to_sys_api_serialized_date(date):
        """
        Converts a datetime object to a string containing a System API-specific representation of a date.
        The implementation is adapted from SysTimestampHelper.cs (search under ni-central repo).

        :param date: The date in datetime format.
        :return: The converted date represented as a string.
        """
        max_ulong = 0xffffffffffffffff
        offset_between_0001_and_1904 = 0xfffffff20493b980
        ticks_conversions_unit = 10000000

        td = date - datetime(1, 1, 1, tzinfo=timezone.utc)
        ticks = int((td / timedelta(microseconds=1)) * 10)

        sec_since_epoch_0001 = int(ticks / ticks_conversions_unit)
        sec_since_epoch_1904 = sec_since_epoch_0001 + offset_between_0001_and_1904

        # System API relies on the rollover property of unsigned int.
        # Since we don't have an unsigned int type readily available,
        # we'll handle the rollover manually.
        if sec_since_epoch_1904 > max_ulong:
            sec_since_epoch_1904 -= max_ulong + 1

        ticks -= sec_since_epoch_0001 * ticks_conversions_unit
        frac_real = float(ticks) / ticks_conversions_unit
        frac_int = int(2.0 ** (-64.0) * frac_real)

        t_4 = sec_since_epoch_1904 >> 32
        t_3 = sec_since_epoch_1904
        t_2 = frac_int >> 32
        t_1 = frac_int

        return '{0} {1} {2} {3}'.format(format(t_1, 'X'), format(t_2, 'X'), format(t_3, 'X'), format(t_4, 'X'))
