import fnmatch
import json
import os
from pathlib import Path

from asset_discovery.repository.json_provider import DataProvider
from asset_discovery.repository.models.json_data import JsonData


class FileJsonProvider(DataProvider):
    def __init__(self, store_directory_path):
        self._store_directory_path = store_directory_path

    def get_all(self):
        file_paths = self._get_json_file_paths(self._store_directory_path)
        return self._get_json_file_contents(file_paths)

    @staticmethod
    def _get_json_file_paths(store_directory_path):
        """
        Looks at the _store_directory_path, and identifies all JSON files, returning their absolute paths.

        :return: The paths of all JSON files under the UserDefined directory.
        """
        file_paths = []

        for root, _, filenames in os.walk(store_directory_path):
            files_in_root = [os.path.join(root, filename) for filename in fnmatch.filter(filenames, '*.json')]
            file_paths.extend(files_in_root)

        return file_paths

    @staticmethod
    def _get_json_file_contents(file_paths):
        """
        Given a list of file paths to JSON files, extracts the contents of the JSON files
        and merges them into a single list.

        :param file_paths: The paths of the JSON files from which to read the content.
        :return: A list JsonContent objects, each containing the properties for a JSON file.
        """
        result = []

        for file_path in file_paths:
            try:
                with open(file_path) as json_file:
                    content = json.load(json_file)
                if content:
                    json_data = JsonData(file_path, content)
                    result.append(json_data)
            except (IOError, ValueError):
                pass  # TODO

        return result

    def create(self, json_data):
        file_path = json_data.uri
        try:
            self._ensure_store_directory_path_exists()
            with open(file_path, 'x') as json_file:
                json.dump(json_data.content, json_file, sort_keys=True, indent=4)
        except (IOError, ValueError):
            pass  # TODO

    def _ensure_store_directory_path_exists(self):
        directory_path = os.path.join(self._store_directory_path, "_server")
        Path(directory_path).mkdir(parents=True, exist_ok=True)

    def replace(self, json_data):
        file_path = json_data.uri
        try:
            self._ensure_store_directory_path_exists()
            with open(file_path, 'w') as json_file:
                json.dump(json_data.content, json_file, sort_keys=True, indent=4)
        except (IOError, ValueError):
            pass  # TODO

    def remove(self, json_uri):
        try:
            os.remove(json_uri)
        except (IOError, ValueError):
            pass  # TODO

    def get_qualified_uri(self, base_uri):
        file_name = "{base_uri}.json".format(base_uri=base_uri)
        file_path = os.path.join(self._store_directory_path, "_server", file_name)
        return file_path
