import uuid
from typing import Optional

from asset_discovery.models import Location, Defaults, AssetIdentification, Asset
from asset_discovery.exceptions import DuplicateError, NotFoundError
from asset_discovery.repository.data_store import DataStore
from asset_discovery.repository.models.json_data import JsonData
from asset_discovery.repository.in_memory_entry_map import InMemoryEntryMap
from asset_discovery.repository.file_json_provider import FileJsonProvider
from asset_discovery.repository.json_provider import DataProvider
from asset_discovery.util.asset_serializer import AssetSerializer
from asset_discovery.util.object_patcher import ObjectPatcher

DEFAULT = Defaults.DEFAULT_VALUE


class JsonDataStore(DataStore):
    def __init__(self, json_provider: DataProvider = None, serializer=None, top_level_property_name="assets"):
        if json_provider is None:
            json_provider = FileJsonProvider()
        super().__init__(json_provider)

        if serializer is None:
            serializer = AssetSerializer()
        self._entry_serializer = serializer
        self._top_level_property_name = top_level_property_name
        self._in_memory_entry_map = InMemoryEntryMap()

    def list(self):
        self._refresh_entry_map()
        return self._in_memory_entry_map.get_all()

    def get_by_identification(self, asset_identification: AssetIdentification) -> Optional[Asset]:
        self._refresh_entry_map()
        return self._in_memory_entry_map.get_by_identification(asset_identification)

    def get_identification_to_entry(self):
        self._refresh_entry_map()
        return self._in_memory_entry_map.get_identification_to_entry()

    def _refresh_entry_map(self):
        self._in_memory_entry_map.flush()
        json_data_result = self._json_provider.get_all()
        for json_data in json_data_result:
            serialized_entries = json_data.content.get(self._top_level_property_name, None)
            if not isinstance(serialized_entries, list):
                continue
            for serialized_entry in serialized_entries:
                if not isinstance(serialized_entry, dict):
                    continue
                entry = self._entry_serializer.from_json(serialized_entry)
                if entry is not None:
                    self._in_memory_entry_map.add(json_data.uri, entry)

    def add(self, entry):
        self._refresh_entry_map()
        exists = self._in_memory_entry_map.exists(entry.identification)
        if exists:
            raise DuplicateError(entry.identification)

        uri = self._generate_uri()

        self._create_json_with_entry(uri, entry)

    def _create_json_with_entry(self, uri, entry):
        serialized_entry = self._entry_serializer.to_json(entry)
        entries_json = {self._top_level_property_name: [serialized_entry]}
        self._json_provider.create(JsonData(uri, entries_json))

    def _generate_uri(self):
        base_uri = uuid.uuid4()
        return self._json_provider.get_qualified_uri(base_uri)

    def remove(self, identification):
        self._refresh_entry_map()
        existing_entry = self._in_memory_entry_map.get_by_identification(identification)
        if existing_entry is None:
            raise NotFoundError(identification)

        uri = self._in_memory_entry_map.get_uri(identification)
        bundled_entries = self._in_memory_entry_map.get_by_uri(uri)

        entries = [a for a in bundled_entries if a.identification != identification]

        if not bool(entries):
            self._json_provider.remove(uri)
            return

        self._replace_json_with_entries(uri, entries)

    def _replace_json_with_entries(self, uri, entries):
        serialized_entries = [self._entry_serializer.to_json(entry) for entry in entries]
        entries_json = {self._top_level_property_name: serialized_entries}
        self._json_provider.replace(JsonData(uri, entries_json))

    def remove_all(self):
        self._refresh_entry_map()
        existing_entries = self._in_memory_entry_map.get_all()
        existing_uris = {self._in_memory_entry_map.get_uri(entry.identification) for entry in existing_entries}

        for uri in existing_uris:
            self._json_provider.remove(uri)

    def update(self, update_data, source_props_only=False):
        self._refresh_entry_map()
        existing_entry = self._in_memory_entry_map.get_by_identification(update_data.identification)
        if existing_entry is None:
            raise NotFoundError(update_data.identification)

        updated_entry = self._get_updated_entry(existing_entry, update_data, source_props_only)

        uri = self._in_memory_entry_map.get_uri(updated_entry.identification)
        bundled_entries = self._in_memory_entry_map.get_by_uri(uri)
        entries = [a if a.identification != updated_entry.identification else updated_entry for a in bundled_entries]

        self._replace_json_with_entries(uri, entries)

    @staticmethod
    def _get_updated_entry(entry, update_data, source_props_only):
        if update_data.location is not None and update_data.location is not DEFAULT:
            if entry.location is None:
                entry.location = Location()
            entry.location = ObjectPatcher.patch(entry.location, update_data.location, source_props_only, DEFAULT)
            update_data.location = entry.location
        return ObjectPatcher.patch(entry, update_data, source_props_only, DEFAULT)
