import os
from asset_discovery.environment import Environment
try:
    import winreg
except ImportError:
    pass  # We are on Linux, this is not an issue


class PathResolver:
    NI_INSTALLERS_REG_PATH = 'SOFTWARE\\National Instruments\\Common\\Installer'
    NI_INSTALLERS_REG_KEY_APP_DATA = 'NIPUBAPPDATADIR'
    SKYLINE_DIRECTORY = 'Skyline'

    @staticmethod
    def get_joined_skyline_appdata_dir(*args):
        """
        Return the app data directory path joined with the given args.

        :return: The path of the UserDefined directory.
        """
        common_appdata_dir = PathResolver._get_skyline_appdata_dir()
        user_defined_directory_path = os.path.join(common_appdata_dir, *args)

        return user_defined_directory_path

    @staticmethod
    def _get_skyline_appdata_dir():
        """
        Gets the Skyline AppData directory based on the current platform.

        :return: The path of the Skyline AppData directory.
        """
        if os.name == "nt":
            return PathResolver._get_windows_skyline_appdata_dir()
        return PathResolver._get_linux_skyline_appdata_dir()

    @staticmethod
    def _get_windows_skyline_appdata_dir():
        """
        Gets the Skyline AppData directory on Windows.

        This looks like: 'C:\\ProgramData\\National Instruments\\Skyline'

        :return: The path of the Skyline AppData directory.
        """
        if Environment.DEBUG_MODE:
            # TODO: Add logic to get this through an an env variable.
            return r'C:\ProgramData\National Instruments\Skyline'
        with winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                PathResolver.NI_INSTALLERS_REG_PATH,
                0,
                winreg.KEY_READ) as hkey:
            (appdata_dir, _) = winreg.QueryValueEx(
                hkey, PathResolver.NI_INSTALLERS_REG_KEY_APP_DATA)
            return os.path.join(
                appdata_dir,
                PathResolver.SKYLINE_DIRECTORY
            )

    @staticmethod
    def _get_linux_skyline_appdata_dir():
        """
        Gets the Skyline AppData directory on Windows.

        This looks like: '/etc/natinst/niskyline'

        :return: The path of the Skyline AppData directory.
        """
        return '/etc/natinst/niskyline'
