import datetime
import logging
from typing import Optional, Any, Dict, List

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from asset_discovery.repository.entry_serializer import EntrySerializer
from asset_discovery.models import (
    Asset,
    BusType,
    AssetType,
    AssetPresenceStatus,
    Location,
    AssetPresence,
    SelfCalibration,
    ExternalCalibration,
    TemperatureSensor,
    CalibrationMode,
    CalibrationOperator,
    AssetHealthData,
    LifetimePowerOn,
    OperationalState,
    FanSensorModel,
    SensorCollection,
    SensorModel,
    CoolingHealth,
    SensorHealthState)
from asset_discovery.util import asset_serializer_constants

logger = logging.getLogger()


class AssetSerializer(EntrySerializer):
    BUS_TYPE_VALUES = {
        'BuiltInSystem': BusType.BUILT_IN_SYSTEM,
        'BUILT_IN_SYSTEM': BusType.BUILT_IN_SYSTEM,
        'PciPxi': BusType.PCI_PXI,
        'PCI_PXI': BusType.PCI_PXI,
        'Usb': BusType.USB,
        'USB': BusType.USB,
        'Gpib': BusType.GPIB,
        'GPIB': BusType.GPIB,
        'Vxi': BusType.VXI,
        'VXI': BusType.VXI,
        'Serial': BusType.SERIAL,
        'SERIAL': BusType.SERIAL,
        'TcpIp': BusType.TCP_IP,
        'TCP_IP': BusType.TCP_IP,
        'cRio': BusType.CRIO,
        'CRIO': BusType.CRIO,
        'Scxi': BusType.SCXI,
        'SCXI': BusType.SCXI,
        'cDaq': BusType.CDAQ,
        'CDAQ': BusType.CDAQ,
        'SwitchBlock': BusType.SWITCH_BLOCK,
        'SWITCH_BLOCK': BusType.SWITCH_BLOCK,
        'Scc': BusType.SCC,
        'SCC': BusType.SCC,
        'FireWire': BusType.FIRE_WIRE,
        'FIRE_WIRE': BusType.FIRE_WIRE,
        'Accessory': BusType.ACCESSORY,
        'ACCESSORY': BusType.ACCESSORY,
        'Can': BusType.CAN,
        'CAN': BusType.CAN,
        'SwitchBlockDevice': BusType.SWITCH_BLOCK_DEVICE,
        'SWITCH_BLOCK_DEVICE': BusType.SWITCH_BLOCK_DEVICE
    }

    ASSET_TYPE_VALUES = {
        'Generic': AssetType.GENERIC,
        'GENERIC': AssetType.GENERIC,
        'DeviceUnderTest': AssetType.DEVICE_UNDER_TEST,
        'DEVICE_UNDER_TEST': AssetType.DEVICE_UNDER_TEST,
        'Fixture': AssetType.FIXTURE,
        'FIXTURE': AssetType.FIXTURE,
        'System': AssetType.SYSTEM,
        'SYSTEM': AssetType.SYSTEM,
        'Chassis': AssetType.CHASSIS,
        'CHASSIS': AssetType.CHASSIS
    }

    ASSET_PRESENCE_VALUES = {
        'Initializing': AssetPresenceStatus.INITIALIZING,
        'INITIALIZING': AssetPresenceStatus.INITIALIZING,
        'Unknown': AssetPresenceStatus.UNKNOWN,
        'UNKNOWN': AssetPresenceStatus.UNKNOWN,
        'NotPresent': AssetPresenceStatus.NOT_PRESENT,
        'NOT_PRESENT': AssetPresenceStatus.NOT_PRESENT,
        'Present': AssetPresenceStatus.PRESENT,
        'PRESENT': AssetPresenceStatus.PRESENT,
    }

    SENSOR_HEALTH_STATE_VALUES = {
        "UNKNOWN": SensorHealthState.UNKNOWN,
        "NORMAL": SensorHealthState.NORMAL,
        "EXCEEDS_USER_DEFINED_THRESHOLD": SensorHealthState.EXCEEDS_USER_DEFINED_THRESHOLD,
        "OUT_OF_SPEC": SensorHealthState.OUT_OF_SPEC,
        "RISK_OF_DAMAGE": SensorHealthState.RISK_OF_DAMAGE
    }

    # TODO: Update JSON serializer

    def __init__(self, infer_built_in_bus_type=False, infer_present=False, infer_empty_temperature_sensors=False):
        self._infer_built_in_bus_type = infer_built_in_bus_type
        self._infer_present = infer_present
        self._infer_empty_temperature_sensors = infer_empty_temperature_sensors

    def from_json(self, json_content, partial_model=False):
        """
        Given a list of asset properties, builds a list of AssetUpdateModel objects.

        :param json_content: A dictionary containing the properties for an asset.
        :param partial_model: Flag used to specify whether asset is partial, meaning it will not have some core fields,
        such as identification.
        :return: An Asset object representing the deserialized asset.
        """
        properties = json_content.copy()
        asset = self._create_asset_with_identification_properties(properties)
        if asset is None and partial_model:
            # we add network devices by specifying their ip or hostname, without knowing their identifications
            asset = Asset()
        if asset:
            self._set_health_data_on_asset(properties, asset)
            self._set_location_properties_on_asset(properties, asset)
            self._set_self_calibration_properties_on_asset(properties, asset)
            self._set_external_calibration_properties_on_asset(properties, asset)

            self._set_handled_properties_on_asset(properties, asset)
        return asset

    @classmethod
    def _islist(cls, data):
        """
        Check whether data is a list.

        :param data: The value to be checked.
        :return: ``True`` if data is a list, else ``False``.
        """
        return cls._isinstance(data, list, False)

    @classmethod
    def _isinstance(cls, data, expected_type, inheritance_lookup=True):
        """
        This is our own implementation of isinstance(), since in some cases
        we don't want to take inheritance into account when parsing data.

        For example: `isinstance(False, int)` returns True, but we want to make
        sure the user specified the exact type we expect, which is int. Checking
        `type(False) is int` returns False.

        When you call this as you would call the Python isinstance() function,
        this will behave the same. The inheritance_lookup parameter is set
        by default to True, so it will use the isinstance() function.

        When you call this by specifying the inheritance_lookup parameter
        to False, this will check the data's exact type against the
        expected_type passed in.

        :param data: The value to be checked.
        :param expected_type: The type to verify against which to verify data.
        :param inheritance_lookup: Whether to take inheritance into account.
        :return: ``True`` if data is of the specified type, else ``False``.
        """
        if inheritance_lookup:
            return isinstance(data, expected_type)

        return type(data) is expected_type

    @classmethod
    def _isdict(cls, data):
        """
        Check whether data is a dict.

        :param data: The value to be checked.
        :return: ``True`` if data is a dict, else ``False``.
        """
        return cls._isinstance(data, dict, False)

    @classmethod
    def _create_asset_with_identification_properties(cls, properties):
        """
        Extracts and validates identification properties from the bag returning an Asset
        with the identification properties set.

        If any of the properties are not present in the bag or are invalid returns None.

        :param properties: A dictionary containing the properties for an asset.
        :return: The AssetUpdateModel object representing the asset.
        """
        model_number, model_name, has_model = cls._get_number_and_name_paired_properties(
            properties, asset_serializer_constants.MODEL_NUMBER_KEY, asset_serializer_constants.MODEL_NAME_KEY
        )

        vendor_number, vendor_name, has_vendor = cls._get_number_and_name_paired_properties(
            properties, asset_serializer_constants.VENDOR_NUMBER_KEY, asset_serializer_constants.VENDOR_NAME_KEY
        )

        serial_number = properties.get(asset_serializer_constants.SERIAL_NUMBER_KEY, None)
        if not cls._isstr(serial_number):
            serial_number = None

        if not cls._is_asset_valid(has_model, has_vendor):
            return None

        return Asset(
            model_name=model_name,
            model_number=model_number,
            serial_number=serial_number,
            vendor_name=vendor_name,
            vendor_number=vendor_number
        )

    @classmethod
    def _get_number_and_name_paired_properties(cls,
                                               properties,
                                               number_key,
                                               name_key):
        """
        Extracts the values for the given number_key and name_key from the properties and validates them,
        returning both values and a bool indicating success. The operation is considered successful
        if at least one of the properties exists and has a valid value.

        :param properties: A dictionary containing the properties for an asset.
        :param properties: The key of the number property.
        :param properties: The key of the name property.
        :return: A tuple containing the number and name property values and a flag indicating success.
        """
        number = properties.get(number_key, None)
        if not cls._isint(number):
            number = None

        name = properties.get(name_key, None)
        if not cls._isstr(name):
            name = None

        has_paired_properties = number is not None or name is not None

        return number, name, has_paired_properties

    @classmethod
    def _isint(cls, data):
        """
        Check whether data is an int.

        :param data: The value to be checked.
        :return: ``True`` if data is an int, else ``False``.
        """
        return cls._isinstance(data, int, False)

    @classmethod
    def _isstr(cls, data):
        """
        Check whether data is a str.

        :param data: The value to be checked.
        :return: ``True`` if data is a str, else ``False``.
        """
        return cls._isinstance(data, str, False)

    @classmethod
    def _is_asset_valid(cls, has_model, has_vendor):
        """
        Checks that all identification properties exist.

        :param has_model: Whether there are valid model properties.
        :param has_vendor: Whether there are valid vendor properties.
        :return: ``True`` if the asset is valid, else ``False``.
        """
        return has_model and has_vendor

    def _set_location_properties_on_asset(self, properties, asset):
        """
        Extracts the location property from the bag and inserts the location properties
        in the properties list of the asset.

        Properties include:
        - resource uri
        - slot number
        - parent

        :param properties: A dictionary containing the properties for an asset.
        :param asset: An AssetUpdateModel object representing the asset.
        """
        location = properties.get(asset_serializer_constants.LOCATION_KEY, None)
        if not self._isdict(location):
            if self._infer_present:
                asset.location = Location(
                    state=AssetPresence(AssetPresenceStatus.PRESENT)
                )
            return

        resource_uri = location.get(asset_serializer_constants.RESOURCE_URI_KEY, None)

        slot_number = location.get(asset_serializer_constants.SLOT_NUMBER_KEY, None)
        if not self._isint(slot_number):
            slot_number = None

        parent = location.get(asset_serializer_constants.PARENT_KEY, None)

        state_property = location.get(asset_serializer_constants.STATE_KEY, None)
        if not self._infer_present and state_property is None:
            state = state_property
        else:
            # This ensures backwards compatibility. All existing assets from JSON will be considered present
            asset_presence = AssetPresenceStatus.PRESENT
            if self._isdict(state_property):
                asset_presence_property = state_property.get(
                    asset_serializer_constants.ASSET_PRESENCE_KEY, None)
                asset_presence = self.ASSET_PRESENCE_VALUES.get(
                    asset_presence_property,
                    AssetPresenceStatus.UNKNOWN)
            state = AssetPresence(asset_presence=asset_presence)

        asset.location = Location(
            parent=parent,
            resource_uri=resource_uri,
            slot_number=slot_number,
            state=state
        )

    @classmethod
    def _set_self_calibration_properties_on_asset(cls, properties, asset):
        """
        Extracts the self calibration property from the bag and inserts the calibration properties
        in the properties list of the asset.

        Properties include:
        - calibration date
        - limited

        If the calibration date is invalid or if it does not exist this is a NoOp.

        :param properties: A dictionary containing the properties for an asset.
        :param asset: An AssetUpdateModel object representing the asset.
        """
        self_calibration = properties.get(
            asset_serializer_constants.SELF_CALIBRATION_KEY, None)
        if not cls._isdict(self_calibration):
            return

        date_str = self_calibration.get(
            asset_serializer_constants.CALIBRATION_DATE_KEY, None)
        date = cls._parse_date(date_str)
        if not date:
            return

        limited = self_calibration.get(
            asset_serializer_constants.IS_LIMITED_KEY, None)
        if not cls._isbool(limited):
            limited = None

        temp_sensors = self_calibration.get(
            asset_serializer_constants.TEMPERATURE_SENSORS, None)
        temp_sensors = cls._parse_temperature_sensors(temp_sensors, True)

        asset.self_calibration = SelfCalibration(
            temperature_sensors=temp_sensors,
            is_limited=limited,
            date=date
        )

    @classmethod
    def _parse_date(cls, serialized_date):
        """
        Parses a date given as a string into a datetime.

        :param serialized_date: The date represented as a string.
        :return: The parsed date in datetime format.
        """
        try:
            return parse(serialized_date)
        except Exception:  # pylint: disable=broad-except
            pass

        return None

    @classmethod
    def _isbool(cls, data) -> bool:
        """
        Check whether data is a bool.

        :param data: The value to be checked.
        :return: ``True`` if data is a bool, else ``False``.
        """
        return cls._isinstance(data, bool, False)

    @classmethod
    def _set_external_calibration_properties_on_asset(cls, properties, asset):
        """
        Extracts the external calibration property from the bag and inserts the calibration properties
        in the properties list of the asset.

        Properties include:
        - calibration date
        - next recommended date
        - recommended interval
        - limited
        - comments
        - checksum
        - operator

        If either one of the calibration date, recommended interval or next recommended date
        are invalid or if they do not exist in the bag this is a NoOp.

        :param properties: A dictionary containing the properties for an asset.
        :param asset: An AssetUpdateModel object representing the asset.
        """
        external_calibration = properties.get(
            asset_serializer_constants.EXTERNAL_CALIBRATION_KEY, None)
        if not cls._isdict(external_calibration):
            return

        date_str = external_calibration.get(
            asset_serializer_constants.CALIBRATION_DATE_KEY, None)
        date = cls._parse_date(date_str)
        interval = external_calibration.get(
            asset_serializer_constants.RECOMMENDED_INTERVAL_KEY, None)
        if not date or not cls._is_interval_valid(interval):
            return

        next_date_str = external_calibration.get(asset_serializer_constants.NEXT_RECOMMENDED_DATE_KEY, None)
        next_date = cls._parse_date(next_date_str)
        if not next_date:
            next_date = date + relativedelta(months=+interval)

        custom_date_str = external_calibration.get(asset_serializer_constants.NEXT_CUSTOM_DUE_DATE, None)
        custom_date = cls._parse_date(custom_date_str)

        calibration_type = external_calibration.get(
            asset_serializer_constants.CALIBRATION_TYPE, None)
        calibration_type = cls._get_calibration_type_or_default(calibration_type, CalibrationMode.MANUAL)
        comments = external_calibration.get(asset_serializer_constants.COMMENTS_KEY, None)
        checksum = external_calibration.get(asset_serializer_constants.CHECKSUM_KEY, None)

        temp_sensors = external_calibration.get(
            asset_serializer_constants.TEMPERATURE_SENSORS, None)
        temp_sensors = cls._parse_temperature_sensors(temp_sensors, True)

        limited = external_calibration.get(
            asset_serializer_constants.IS_LIMITED_KEY, None)
        if not cls._isbool(limited):
            limited = None

        operator = None
        operator_property = external_calibration.get(asset_serializer_constants.OPERATOR_KEY, None)
        if cls._isdict(operator_property):
            display_name = operator_property.get(asset_serializer_constants.DISPLAY_NAME_KEY, None)
            user_id = operator_property.get(asset_serializer_constants.USER_ID_KEY, None)
            operator = CalibrationOperator(
                display_name=display_name,
                user_id=user_id
            )

        asset.external_calibration = ExternalCalibration(
            temperature_sensors=temp_sensors,
            is_limited=limited,
            date=date,
            recommended_interval=interval,
            next_recommended_date=next_date,
            next_custom_due_date=custom_date,
            comments=comments,
            entry_type=calibration_type,
            operator=operator,
            checksum=checksum
        )

    @classmethod
    def _is_interval_valid(cls, data):
        """
        Check whether data is a valid calibration interval.

        :param data: The value to be checked.
        :return: ``True`` if data is a valid interval, else ``False``.
        """
        return cls._isint(data) and int(data) > 0

    @classmethod
    def _get_calibration_type_or_default(cls, calibration_type, default):
        """
        Determine whether the calibration type passed as a parameter is valid.
        If valid, return the value, otherwise return the default value.
        In order for a calibration type to be valid it must be equal either to
        MANUAL or AUTOMATIC (case sensitive).

        :param calibration_type: Nullable value for the calibration type.
        :param default: Default value for the calibration type.
        :return: The calibration type passed as a parameter if the value is valid
        else the default value.
        """
        if calibration_type in [CalibrationMode.MANUAL, CalibrationMode.AUTOMATIC]:
            return calibration_type
        return default

    @classmethod
    def _parse_temperature_sensors(cls, sensors, infer_empty_temperature_sensors):
        """
        Parses a list of dictionaries to a list of temperature sensor models.

        :param sensors: A list of dictionaries containing temperature sensors data.
        :return: Empty list if the sensors list is NIL, the parsed list otherwise.
        """
        if not cls._islist(sensors):
            temp_sensors = []
        else:
            temp_sensors = [sm for sm in [cls._parse_temperature_sensor(s) for s in sensors] if sm is not None]
        if not infer_empty_temperature_sensors and not temp_sensors:
            return None
        else:
            return temp_sensors

    @classmethod
    def _parse_temperature_sensor(cls, sensor):
        """
        Parses a dictionary to a TemperatureSensorModel.
        A temperature sensor model is only valid if the reading is a non null number.

        :param sensor: A dictionary containing property of a temperature sensor model.
        :return: The parsed model or none if the data is invalid.
        """
        reading = sensor.get(asset_serializer_constants.SENSOR_READING, None)
        if not cls._isint(reading) and not cls._isinstance(reading, float, False):
            return None
        name = sensor.get(asset_serializer_constants.SENSOR_NAME, None)
        return TemperatureSensor(name=name, reading=reading)

    def _set_handled_properties_on_asset(self, properties, asset):
        """
        Extracts all handled properties from the bag (except for identification, asset class and self/external calibration)
        and inserts them in the properties list of the asset.

        :param properties: A dictionary containing the properties for an asset.
        :param asset: An AssetUpdateModel object representing the asset.
        """
        asset_type_property = properties.get(asset_serializer_constants.ASSET_TYPE_KEY, None)
        asset_type = self.ASSET_TYPE_VALUES.get(asset_type_property, None)

        bus_type_property = properties.get(asset_serializer_constants.BUS_TYPE_KEY, None)
        bus_type = self.BUS_TYPE_VALUES.get(
            bus_type_property,
            BusType.BUILT_IN_SYSTEM if self._infer_built_in_bus_type else None)

        provided_bus_type_property = properties.get(asset_serializer_constants.PROVIDED_BUS_TYPE_KEY)
        provided_bus_type = self.BUS_TYPE_VALUES.get(provided_bus_type_property, None)

        asset_name = properties.get(asset_serializer_constants.NAME_KEY, None)
        if asset_name is None:
            asset_name = properties.get(asset_serializer_constants.ASSET_NAME_KEY, None)
        firmware_version = properties.get(asset_serializer_constants.FIRMWARE_VERSION_KEY, None)
        hardware_version = properties.get(asset_serializer_constants.HARDWARE_VERSION_KEY, None)
        visa_resource_name = properties.get(asset_serializer_constants.VISA_RESOURCE_NAME_KEY, None)

        is_ni_asset = properties.get(asset_serializer_constants.IS_NI_ASSET_KEY, None)
        if not self._isbool(is_ni_asset):
            is_ni_asset = None

        supports_self_calibration = properties.get(
            asset_serializer_constants.SUPPORTS_SELF_CALIBRATION_KEY, None)
        if not self._isbool(supports_self_calibration):
            supports_self_calibration = None

        supports_external_calibration = properties.get(
            asset_serializer_constants.SUPPORTS_EXTERNAL_CALIBRATION_KEY, None)
        if not self._isbool(supports_external_calibration):
            supports_external_calibration = None

        supports_self_test = properties.get(
            asset_serializer_constants.SUPPORTS_SELF_TEST_KEY, None)
        if not self._isbool(supports_self_test):
            supports_self_test = None

        supports_reset = properties.get(
            asset_serializer_constants.SUPPORTS_RESET_KEY, None)
        if not self._isbool(supports_reset):
            supports_reset = None

        temp_sensors = properties.get(
            asset_serializer_constants.TEMPERATURE_SENSORS, None)
        temp_sensors = self._parse_temperature_sensors(temp_sensors, self._infer_empty_temperature_sensors)

        ip_addresses = properties.get(asset_serializer_constants.IP_ADDRESSES_KEY, None)
        hostname = properties.get(asset_serializer_constants.HOSTNAME_KEY, None)

        custom_properties = properties.get(asset_serializer_constants.PROPERTIES_KEY, None)
        if not self._are_custom_properties_valid(custom_properties):
            custom_properties = None

        custom_keywords = properties.get(asset_serializer_constants.KEYWORDS_KEY, None)
        if not self._are_custom_keywords_valid(custom_keywords):
            custom_keywords = None

        asset.name = asset_name
        asset.asset_type = asset_type
        asset.bus_type = bus_type
        asset.provided_bus_type = provided_bus_type
        asset.firmware_version = firmware_version
        asset.hardware_version = hardware_version
        asset.visa_resource_name = visa_resource_name
        asset.is_ni_asset = is_ni_asset
        asset.supports_self_calibration = supports_self_calibration
        asset.supports_external_calibration = supports_external_calibration
        asset.supports_self_test = supports_self_test
        asset.supports_reset = supports_reset
        asset.temperature_sensors = temp_sensors
        asset.ip_addresses = ip_addresses
        asset.hostname = hostname
        asset.properties = custom_properties
        asset.keywords = custom_keywords

    @classmethod
    def _are_custom_properties_valid(cls, custom_properties):
        if not cls._isdict(custom_properties):
            return False
        return all(cls._isstr(k) and cls._isstr(v) for k, v in custom_properties.items())

    @classmethod
    def _are_custom_keywords_valid(cls, custom_keywords):
        if not cls._islist(custom_keywords):
            return False
        return all(cls._isstr(k) for k in custom_keywords)

    def to_json(self, asset, use_raw_keys=False):
        return self._sanitize_for_serialization(asset, use_raw_keys)

    @staticmethod
    def _sanitize_for_serialization(obj, use_raw_keys):
        try:
            if obj is None:
                return None
            elif isinstance(obj, (float, bool, bytes, str, int)):
                return obj
            elif isinstance(obj, list):
                return [AssetSerializer._sanitize_for_serialization(sub_obj, use_raw_keys)
                        for sub_obj in obj]
            elif isinstance(obj, tuple):
                return tuple(AssetSerializer._sanitize_for_serialization(sub_obj, use_raw_keys)
                             for sub_obj in obj)
            elif isinstance(obj, (datetime.datetime, datetime.date)):
                return obj.isoformat()

            if isinstance(obj, dict):
                obj_dict = obj
            else:
                # Convert model obj to dict except
                # attributes `property_types`, `serialization_keys`
                # and attributes which value is not None.
                # Convert attribute name to json key if use_raw_keys is False.
                obj_dict = {attr if use_raw_keys else obj.serialization_keys[attr]: getattr(obj, attr)
                            for attr, _ in obj.property_types.items()
                            if getattr(obj, attr) is not None}

            return {key: AssetSerializer._sanitize_for_serialization(val, use_raw_keys)
                    for key, val in obj_dict.items()}
        except Exception as e:
            logger.error(e)
            logger.error("Asset serialization failed in 'asset_serializer', method '_sanitize_for_serialization'")
            raise e

    @classmethod
    def _set_health_data_on_asset(cls, properties: Dict[str, Any], asset: Asset) -> None:
        health = properties.get(asset_serializer_constants.HEALTH_KEY, None)
        if health is None:
            return None

        temperature_sensors = health.get(asset_serializer_constants.HEALTH_TEMPERATURE_SENSORS_KEY, None)
        temperature_sensors = cls._parse_sensors(temperature_sensors)

        voltage_sensors = health.get(asset_serializer_constants.HEALTH_VOLTAGE_SENSORS_KEY, None)
        voltage_sensors = cls._parse_sensors(voltage_sensors)

        power_sensors = health.get(asset_serializer_constants.HEALTH_POWER_SENSORS_KEY, None)
        power_sensors = cls._parse_sensors(power_sensors)

        lifetime_power_on = health.get(asset_serializer_constants.HEALTH_LIFETIME_POWER_ON_KEY, None)
        lifetime_power_on = cls._parse_lifetime_power_on(lifetime_power_on)

        cooling_health = health.get(asset_serializer_constants.COOLING_HEALTH_KEY, None)
        cooling_health = cls._parse_cooling_health(cooling_health)

        operational_state = health.get(asset_serializer_constants.HEALTH_OPERATIONAL_STATE_KEY, None)
        operational_state = cls._parse_operational_state(operational_state)

        if (not temperature_sensors and not voltage_sensors and not power_sensors and
                not lifetime_power_on and not cooling_health and not operational_state):
            asset.health = None
            return

        health = AssetHealthData(
            temperature_sensors=temperature_sensors,
            voltage_sensors=voltage_sensors,
            power_sensors=power_sensors,
            lifetime_power_on=lifetime_power_on,
            cooling_health=cooling_health,
            operational_state=operational_state
        )
        asset.health = health

    @classmethod
    def _parse_cooling_health(cls, raw_cooling_health: Dict[str, Any]) -> Optional[CoolingHealth]:
        if not cls._isdict(raw_cooling_health):
            return None

        timestamp = raw_cooling_health.get(asset_serializer_constants.HEALTH_TIMESTAMP_KEY, None)
        timestamp = cls._parse_date(timestamp)

        fan_mode = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_FAN_MODE_KEY, None)
        supported_fan_modes = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_SUPPORTED_FAN_MODES_KEY, None)
        user_desired_fan_speed = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_USER_DESIRED_FAN_SPEED_KEY, None)
        manual_rpm_upper_bound = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_MANUAL_RPM_UPPER_BOUND_KEY, None)
        manual_rpm_lower_bound = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_MANUAL_RPM_LOWER_BOUND_KEY, None)
        cooling_profile = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_COOLING_PROFILE_KEY, None)
        supported_cooling_profiles = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_SUPPORTED_COOLING_PROFILES_KEY, None)
        cooling_operational_state = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_COOLING_OPERATIONAL_STATE_KEY, None)

        chassis_fans = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_CHASSIS_FAN_SENSORS_KEY, None)
        chassis_fans = cls._parse_fan_sensors(chassis_fans)

        power_supply_fans = raw_cooling_health.get(asset_serializer_constants.COOLING_HEALTH_POWER_SUPPLY_FAN_SENSORS_KEY, None)
        power_supply_fans = cls._parse_power_supply_fan_sensors(power_supply_fans)

        if (not fan_mode and not supported_fan_modes and not user_desired_fan_speed and not manual_rpm_upper_bound and
                not manual_rpm_lower_bound and not cooling_profile and not supported_cooling_profiles and
                not cooling_operational_state and not chassis_fans and not power_supply_fans):
            return None
        raw_cooling_health = CoolingHealth(
            fan_mode=fan_mode,
            supported_fan_modes=supported_fan_modes,
            user_desired_fan_speed=user_desired_fan_speed,
            manual_rpm_upper_bound=manual_rpm_upper_bound,
            manual_rpm_lower_bound=manual_rpm_lower_bound,
            cooling_profile=cooling_profile,
            supported_cooling_profiles=supported_cooling_profiles,
            cooling_operational_state=cooling_operational_state,
            chassis_fans=chassis_fans,
            power_supply_fans=power_supply_fans,
            timestamp=timestamp)
        return raw_cooling_health

    @classmethod
    def _parse_fan_sensors(cls, raw_sensors_list: Dict[str, Any]) -> Optional[List[FanSensorModel]]:
        if not cls._islist(raw_sensors_list):
            return None

        fan_sensor_models = []
        for raw_sensor in raw_sensors_list:
            fan_sensor_models.append(cls._parse_fan_sensor(raw_sensor))

        return fan_sensor_models

    @classmethod
    def _parse_fan_sensor(cls, raw_sensor: Dict[str, Any]) -> FanSensorModel:
        reading = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_READING_KEY, None)
        name = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_NAME_KEY, None)
        raw_state = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_STATE_KEY, None)
        state = cls.SENSOR_HEALTH_STATE_VALUES.get(raw_state, None)
        properties = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_PROPERTIES_KEY, None)

        return FanSensorModel(
            name=name,
            reading=reading,
            state=state,
            properties=properties
        )

    @classmethod
    def _parse_power_supply_fan_sensors(cls, raw_sensors_list: Dict[str, Any]) -> Optional[List[FanSensorModel]]:
        if not cls._islist(raw_sensors_list):
            return None

        fan_sensor_models = []
        for raw_sensor in raw_sensors_list:
            sensor = FanSensorModel()
            sensor.reading = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_READING_KEY, None)
            sensor.name = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_NAME_KEY, None)
            fan_sensor_models.append(sensor)

        return fan_sensor_models

    @classmethod
    def _parse_sensors(cls, raw_sensors: Dict[str, Any]) -> Optional[SensorCollection]:
        if raw_sensors is None:
            return None
        timestamp = raw_sensors.get(asset_serializer_constants.HEALTH_TIMESTAMP_KEY, None)
        if timestamp is None:
            return None

        raw_sensors_list = raw_sensors.get(asset_serializer_constants.HEALTH_SENSORS_KEY, None)
        if raw_sensors_list is None or not cls._islist(raw_sensors_list):
            return None

        sensor_models = []
        for raw_sensor in raw_sensors_list:
            sensor_models.append(cls._parse_sensor(raw_sensor))

        timestamp = cls._parse_date(timestamp)
        return SensorCollection(timestamp=timestamp, sensors=sensor_models)

    @classmethod
    def _parse_sensor(cls, raw_sensor: Dict[str, Any]) -> SensorModel:
        reading = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_READING_KEY, None)
        sensor_type = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_TYPE_KEY, None)
        name = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_NAME_KEY, None)
        raw_state = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_STATE_KEY, None)
        state = cls.SENSOR_HEALTH_STATE_VALUES.get(raw_state, None)
        raw_alarm = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_ALARM_KEY, None)
        alarm = cls.SENSOR_HEALTH_STATE_VALUES.get(raw_alarm, None)
        properties = raw_sensor.get(asset_serializer_constants.HEALTH_SENSOR_PROPERTIES_KEY, {})
        return SensorModel(
            sensor_type=sensor_type,
            name=name,
            reading=reading,
            state=state,
            alarm=alarm,
            properties=properties
        )

    @classmethod
    def _parse_lifetime_power_on(cls, raw_lifetime_power_on: Dict[str, Any]):
        if not cls._isdict(raw_lifetime_power_on):
            return None
        timestamp = raw_lifetime_power_on.get(asset_serializer_constants.HEALTH_TIMESTAMP_KEY, None)
        if timestamp is None:
            return None
        timestamp = cls._parse_date(timestamp)

        hours = raw_lifetime_power_on.get(asset_serializer_constants.HEALTH_POWER_ON_HOURS_KEY, None)
        if not cls._isint(hours):
            return None
        hours = int(hours)

        return LifetimePowerOn(
            power_on_hours=hours,
            timestamp=timestamp
        )

    @classmethod
    def _parse_operational_state(cls, raw_operational_state: Dict[str, Any]):
        if not cls._isdict(raw_operational_state):
            return None
        timestamp = raw_operational_state.get(asset_serializer_constants.HEALTH_TIMESTAMP_KEY, None)
        if timestamp is None:
            return None
        timestamp = cls._parse_date(timestamp)

        state_number = raw_operational_state.get(asset_serializer_constants.HEALTH_OPERATIONAL_STATE_NUMBER_KEY, None)
        if not cls._isint(state_number):
            return None
        state_number = int(state_number)

        return OperationalState(
            state_number=state_number,
            timestamp=timestamp
        )
